/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         stemnt.cpp
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of statements.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#include "stemnt.h"
#include "symbol.h"

#include "gram.h"


// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Label::Label( LabelType labelType )
{
    type = labelType;
    Expr = NULL;
    next = NULL;
}   

Label::Label( Expression *expr )
{
    type = LT_Case;
    Expr = expr;
    next = NULL;
}   

Label::Label( Symbol *sym )
{
    type = LT_Goto;
    Name = sym;
    next = NULL;
}   

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement::Statement(StatementType stemntType)
{
    type  = stemntType;

    level = 0;

    lblHead = lblTail = NULL;
    next = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement::~Statement()
{
    Label    *lbl, *prevLbl=NULL;

    for (lbl=lblHead; lbl; lbl=lbl->next)
    {
        delete prevLbl;
        prevLbl = lbl;
    }
    delete prevLbl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Statement::AddLabel( Label *lbl )
{
    lbl->next = NULL;

    if (lblTail)
        lblTail->next = lbl;
    else
        lblHead = lbl;

    lblTail = lbl;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ExpressionStemnt::ExpressionStemnt( Expression *expr )
                 : Statement( ST_ExpressionStemnt )
{
    expression = expr;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ExpressionStemnt::~ExpressionStemnt()
{
    delete expression;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
IfStemnt::IfStemnt( Expression *_cond,
                    Statement *_thenBlk,
                    Statement *_elseBlk /* =NULL */)
         : Statement( ST_IfStemnt )
{
    cond = _cond;
    thenBlk = _thenBlk;
    elseBlk = _elseBlk;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
IfStemnt::~IfStemnt()
{
    delete cond;
    delete thenBlk;
    delete elseBlk;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
SwitchStemnt::SwitchStemnt( Expression *_cond, Statement *_block )
             : Statement( ST_SwitchStemnt )
{
    cond = _cond;
    block = _block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
SwitchStemnt::~SwitchStemnt()
{
    delete cond;
    delete block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ForStemnt::ForStemnt( Expression *_init, Expression *_cond,
                      Expression *_incr, Statement *_block /* =NULL */)
          : Statement( ST_ForStemnt )
{
    init = _init;
    cond = _cond;
    incr = _incr;

    block = _block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ForStemnt::~ForStemnt()
{
    delete init;
    delete cond;
    delete incr;

    delete block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
WhileStemnt::WhileStemnt( Expression *_cond, Statement *_block )
            : Statement( ST_WhileStemnt )
{
    cond = _cond;
    block = _block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
WhileStemnt::~WhileStemnt()
{
    delete cond;
    delete block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
DoWhileStemnt::DoWhileStemnt( Expression *_cond, Statement *_block )
              : Statement( ST_DoWhileStemnt )
{
    cond = _cond;
    block = _block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
DoWhileStemnt::~DoWhileStemnt()
{
    delete cond;
    delete block;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
GotoStemnt::GotoStemnt( Symbol *_dest )
           : Statement( ST_GotoStemnt )
{
    dest = _dest;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
GotoStemnt::~GotoStemnt()
{
    delete dest;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ReturnStemnt::ReturnStemnt( Expression *_result )
             : Statement( ST_ReturnStemnt )
{
    result = _result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
ReturnStemnt::~ReturnStemnt()
{
    delete result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Block::Block()
      : Statement( ST_Block )
{
    head = tail = NULL;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Block::~Block()
{
    Statement *stemnt, *prevStemnt = NULL;

    for (stemnt=head; stemnt; stemnt=stemnt->next)
    {
        delete prevStemnt;
        prevStemnt = stemnt;
    }
    delete prevStemnt;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Block::Add(Statement *stemnt)
{
    stemnt->next = NULL;

    if (tail)
        tail->next = stemnt;
    else
        head = stemnt;

    tail = stemnt;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
void
Block::AddStatements(treenode *node)
{
    Statement    *stemnt;
    Block        *blk;

    if (node)
    {
        switch (node->which)
        {
            default:
            case NONE_T:
            case LEAF_T:
            case IF_T:
            case FOR_T:
                break;
    
            case NODE_T:
                switch (node->type)
                {
                  default:
                      break;

                  case TN_DECL_LIST:
                  case TN_STEMNT_LIST:
                      AddStatements(node->lnode);
                      AddStatements(node->rnode);
                      break;

                  case TN_BLOCK:
                      blk = new Block();
                      blk->AddStatements(node->lnode);
                      blk->AddStatements(node->rnode);
                      Add(blk);
                      break;

                  case TN_DECL:
                  case TN_DECLS:
                  case TN_STEMNT:
                      stemnt = buildStatement(node);
                      Add(stemnt);
                      break;
                }
                break;
        }
    }
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Label*
buildLabel(treenode *node)
{
    Label       *result = NULL;
    Expression  *expr;
    Symbol      *sym;
    leafnode    *leaf;

    if (node)
    {
        switch (node->which)
        {
            default:
            case NONE_T:
            case IF_T:
            case FOR_T:
                break;

            case LEAF_T:
                leaf = (leafnode*) node;
                switch (leaf->tok)
                {
                    default:
                        break;

                    case IDENT:       // goto label
                        // NYI : IDENT -> Symbol
                        sym = buildSymbol((leafnode*) node);
                        result = new Label(sym);
                        break;

                    case DEFLT:       // default
                        result = new Label(LT_Default);
                        break;
                }
                break;
    
            case NODE_T:
                switch (node->tok)
                {
                    default:
                        break;

                    case CASE:        // case label
                        expr = buildExpr(node->rnode);
                        result = new Label(expr);
                        break;
                }
                break;
        }
    }

    return result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Statement*
buildStatement(treenode *node)
{
    Statement    *result = NULL;
    //FunctionDef  *fDef;

    if (node)
    {
        switch (node->which)
        {
            default:
            case NONE_T:
            case LEAF_T:
                break;
    
            case NODE_T:
                switch (node->type)
                {
                  default:
                      break;

                  case TN_DECL:
                      break;

                  case TN_DECLS:
                      break;

                  case TN_BLOCK:
                      {
                      Block *blk = new Block();
                      blk->AddStatements(node->lnode);
                      blk->AddStatements(node->rnode);
                      result = blk;
                      }
                      break;

                  case TN_STEMNT:
                      result = buildStatement(node->rnode);
                      break;

                  case TN_LABEL:
                      {
                      result = buildStatement(node->rnode);
                      Label *lbl = buildLabel(node->lnode); 
                      result->AddLabel(lbl);
                      }
                      break;

                  case TN_EXPR:
                      {
                      Expression *expr = buildExpr(node);  
                      result = new ExpressionStemnt(expr);
                      }
                      break;

                  case TN_SWITCH:
                      {
                      Expression *cond = buildExpr(node->lnode);
                      Statement  *body = buildStatement(node->rnode);
                      result = new SwitchStemnt(cond,body);
                      }
                      break;

                  case TN_JUMP:
                      switch (node->tok)
                      {
                        case BREAK:
                          result = new Statement(ST_BreakStemnt);
                          break;

                        case CONT:
                          result = new Statement(ST_ContinueStemnt);
                          break;

                        case RETURN:
                          {
                          Expression *ret = buildExpr(node->lnode);
                          result = new ReturnStemnt(ret);
                          }
                          break;

                        case GOTO:
                          {
                          // NYI
                          Symbol *dest = buildSymbol((leafnode*) node->lnode);
                          result = new GotoStemnt(dest);
                          }
                          break;
                      }
                      break;

                  case TN_DOWHILE:
                      {
                      Expression *cond = buildExpr(node->lnode);
                      Statement  *body = buildStatement(node->rnode);
                      result = new DoWhileStemnt(cond,body);
                      }
                      break;

                  case TN_WHILE:
                      {
                      Expression *cond = buildExpr(node->lnode);
                      Statement  *body = buildStatement(node->rnode);
                      result = new WhileStemnt(cond,body);
                      }
                      break;
                }
                break;

            case IF_T:
                {
                if_node *ifn = (if_node*) node;

                Expression *cond = buildExpr(ifn->cond);
                Statement  *thenStemnt = buildStatement(ifn->then_n); 
                Statement  *elseStemnt = NULL;

                if (ifn->else_n)
                    elseStemnt = buildStatement(ifn->else_n); 

                result = new IfStemnt(cond,thenStemnt,elseStemnt);
                }
                break;

            case FOR_T:
                {
                for_node *forn = (for_node*) node;

                Expression *init = buildExpr(forn->init);
                Expression *cond = buildExpr(forn->test);
                Expression *incr = buildExpr(forn->incr);
                Statement  *body = buildStatement(forn->stemnt); 

                result = new ForStemnt(init,cond,incr,body);
                }
                break;
        }
    }

    return result;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

