/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         stemnt.h
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of statements.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#ifndef    STEMNT_H
#define    STEMNT_H

#include <stdlib.h>
#include <iostream.h>

#include "symbol.h"
#include "express.h"

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum StatementType
{
    ST_Empty=0,              // The null statement.

    ST_Declaration,
    ST_ExpressionStemnt,

    ST_IfStemnt,
    ST_SwitchStemnt,

    ST_ForStemnt,
    ST_WhileStemnt,
    ST_DoWhileStemnt,

    ST_ContinueStemnt,
    ST_BreakStemnt,
    ST_GotoStemnt,
    ST_ReturnStemnt,

    ST_Block 
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
enum LabelType
{
    LT_None=0,              // No label - invalid.

    LT_Default,             // default:
    LT_Case,                // case <expr>:

    LT_Goto                 // A named label (goto destination).
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Label
{
  public:
    Label( LabelType labelType );

    Label( Expression *expr );
    Label( Symbol *sym );
    
    LabelType    type;

    union
    {
      Expression    *Expr;    // for LT_Case
      Symbol        *Name;    // for LT_Goto
    };

    Label    *next;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

class Statement
{
  public:
    Statement(StatementType stemntType);
    virtual ~Statement();

    void AddLabel( Label *lbl );    // Add a label to this statement.

    StatementType    type;
    int              level;    // The "indent" level.


    Label           *lblHead;  // List of labels.
    Label           *lblTail;

    Statement       *next;     // For making a list of statements.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ExpressionStemnt : public Statement
{
  public:
    ExpressionStemnt( Expression *expr );
    ~ExpressionStemnt();

    Expression    *expression;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class IfStemnt : public Statement
{
  public:
    IfStemnt( Expression *_cond,
              Statement *_thenBlk, Statement *_elseBlk = NULL);
    ~IfStemnt();

    Expression    *cond;

    Statement     *thenBlk;
    Statement     *elseBlk;        // Can be NULL
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class SwitchStemnt : public Statement
{
  public:
    SwitchStemnt( Expression *_cond, Statement *_block );
    ~SwitchStemnt();

    Expression    *cond;

    Statement     *block;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ForStemnt : public Statement
{
  public:
    ForStemnt( Expression *_init, Expression *_cond, Expression *_incr,
               Statement *_block =NULL);
    ~ForStemnt();

    Expression    *init;
    Expression    *cond;
    Expression    *incr;

    Statement     *block;        // Can be NULL
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class WhileStemnt : public Statement
{
  public:
    WhileStemnt( Expression *_cond, Statement *_block );
    ~WhileStemnt();

    Expression    *cond;
    Statement     *block;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class DoWhileStemnt : public Statement
{
  public:
    DoWhileStemnt( Expression *_cond, Statement *_block );
    ~DoWhileStemnt();

    Expression    *cond;
    Statement     *block;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class GotoStemnt : public Statement
{
  public:
    GotoStemnt( Symbol *_dest );
    ~GotoStemnt();

    Symbol        *dest;
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class ReturnStemnt : public Statement
{
  public:
    ReturnStemnt( Expression *_result );
    ~ReturnStemnt();

    Expression    *result;        // Can be NULL.
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

class Declaration : public Statement
{
  public:
    Declaration();
   ~Declaration();
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
class Block : public Statement
{
  public:
    Block();
    virtual ~Block();

    void Add(Statement *stemnt);

    void AddStatements(treenode *node);

    Statement     *head;    // List of statements in this block.
    Statement     *tail;

};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
// Not really true, but this is very convienent.
class FunctionDef : public Block
{
  public:
    FunctionDef();
   ~FunctionDef();
};

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

Label     *buildLabel(treenode *node);
Statement *buildStatement(treenode *node);

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o

#endif  /* STEMNT_H */

